// Copyright 2020 The Jujutsu Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// https://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#![allow(missing_docs)]

use std::collections::BTreeMap;
use std::collections::HashMap;
use std::collections::HashSet;
use std::sync::Arc;

use futures::future::try_join_all;
use pollster::FutureExt as _;

use crate::backend;
use crate::backend::BackendResult;
use crate::backend::TreeId;
use crate::backend::TreeValue;
use crate::repo_path::RepoPath;
use crate::repo_path::RepoPathBuf;
use crate::store::Store;
use crate::tree::Tree;

#[derive(Debug)]
enum Override {
    Tombstone,
    Replace(TreeValue),
}

#[derive(Debug)]
pub struct TreeBuilder {
    store: Arc<Store>,
    base_tree_id: TreeId,
    overrides: BTreeMap<RepoPathBuf, Override>,
}

impl TreeBuilder {
    pub fn new(store: Arc<Store>, base_tree_id: TreeId) -> TreeBuilder {
        let overrides = BTreeMap::new();
        TreeBuilder {
            store,
            base_tree_id,
            overrides,
        }
    }

    pub fn store(&self) -> &Store {
        self.store.as_ref()
    }

    pub fn set(&mut self, path: RepoPathBuf, value: TreeValue) {
        assert!(!path.is_root());
        self.overrides.insert(path, Override::Replace(value));
    }

    pub fn remove(&mut self, path: RepoPathBuf) {
        assert!(!path.is_root());
        self.overrides.insert(path, Override::Tombstone);
    }

    pub fn set_or_remove(&mut self, path: RepoPathBuf, value: Option<TreeValue>) {
        assert!(!path.is_root());
        if let Some(value) = value {
            self.overrides.insert(path, Override::Replace(value));
        } else {
            self.overrides.insert(path, Override::Tombstone);
        }
    }

    pub fn write_tree(self) -> BackendResult<TreeId> {
        if self.overrides.is_empty() {
            return Ok(self.base_tree_id);
        }

        let mut trees_to_write = self.get_base_trees()?;

        // Update entries in parent trees for file overrides
        for (path, file_override) in self.overrides {
            let (dir, basename) = path.split().unwrap();
            let tree = trees_to_write.get_mut(dir).unwrap();
            match file_override {
                Override::Replace(value) => {
                    tree.set(basename.to_owned(), value);
                }
                Override::Tombstone => {
                    tree.remove(basename);
                }
            }
        }

        // All directories to write, for other collections to have references into.
        let dirs: HashSet<RepoPathBuf> = trees_to_write.keys().cloned().collect();
        // Directories we cannot write yet because we need to first write their
        // children.
        let mut blocked_by: HashMap<&RepoPath, HashSet<&RepoPath>> = HashMap::new();
        // Directories we can write next.
        let mut unblocked: HashSet<_> = dirs.iter().map(AsRef::as_ref).collect();
        for dir in &dirs {
            if let Some((parent, _)) = dir.split() {
                blocked_by.entry(parent).or_default().insert(dir);
                unblocked.remove(parent);
            }
        }

        let store = &self.store;
        // On each iteration, write all unblocked trees and find the next set of
        // unblocked directories among the parents of the trees we just wrote.
        while !unblocked.is_empty() {
            if blocked_by.is_empty() {
                // We're writing the root tree. Write it even if empty. Return its id.
                assert_eq!(unblocked.len(), 1);
                let dir = RepoPath::root();
                let tree = trees_to_write.remove(dir).unwrap();
                let written_tree = store.write_tree(&dir, tree).block_on()?;
                return Ok(written_tree.id().clone());
            }

            let mut new_unblocked = HashSet::new();
            let mut writes = vec![];
            let mut dir_writes = vec![];
            for dir in unblocked {
                let tree = trees_to_write.remove(dir).unwrap();
                let (parent, basename) = dir.split().unwrap();
                if tree.is_empty() {
                    let parent_tree = trees_to_write.get_mut(parent).unwrap();
                    if let Some(TreeValue::Tree(_)) = parent_tree.value(basename) {
                        parent_tree.remove(basename);
                    } else {
                        // Entry would have been replaced with file (see
                        // above)
                    }
                } else {
                    writes.push(store.write_tree(&dir, tree));
                    dir_writes.push(dir);
                }

                let parent_blocked_by = blocked_by.get_mut(parent).unwrap();
                parent_blocked_by.remove(dir);
                if parent_blocked_by.is_empty() {
                    blocked_by.remove(parent);
                    new_unblocked.insert(parent);
                }
            }

            let written = try_join_all(writes).block_on()?;
            for (dir, tree) in dir_writes.iter().zip(written) {
                let (parent, basename) = dir.split().unwrap();
                let parent_tree = trees_to_write.get_mut(parent).unwrap();
                parent_tree.set(basename.to_owned(), TreeValue::Tree(tree.id().clone()));
            }

            unblocked = new_unblocked;
        }

        unreachable!("trees_to_write must contain the root tree");
    }

    fn get_base_trees(&self) -> BackendResult<BTreeMap<RepoPathBuf, backend::Tree>> {
        let store = &self.store;
        let mut tree_cache = {
            let dir = RepoPathBuf::root();
            let tree = store.get_tree(dir.clone(), &self.base_tree_id)?;
            BTreeMap::from([(dir, tree)])
        };

        fn populate_trees<'a>(
            tree_cache: &'a mut BTreeMap<RepoPathBuf, Tree>,
            store: &Arc<Store>,
            dir: &RepoPath,
        ) -> BackendResult<&'a Tree> {
            // `if let Some(tree) = ...` doesn't pass lifetime check as of Rust 1.84.0
            if tree_cache.contains_key(dir) {
                return Ok(tree_cache.get(dir).unwrap());
            }
            let (parent, basename) = dir.split().expect("root must be populated");
            let tree = populate_trees(tree_cache, store, parent)?
                .sub_tree(basename)?
                .unwrap_or_else(|| Tree::empty(store.clone(), dir.to_owned()));
            Ok(tree_cache.entry(dir.to_owned()).or_insert(tree))
        }

        for path in self.overrides.keys() {
            let parent = path.parent().unwrap();
            populate_trees(&mut tree_cache, store, parent)?;
        }

        Ok(tree_cache
            .into_iter()
            .map(|(dir, tree)| (dir, tree.data().clone()))
            .collect())
    }
}
